/*************************************************************************/
/**
 * \file GEO-L1SD.c
 * \brief Lee bandas de satélite y genera ficheros temporales binarios y netCDF
 *
 *
 * \par Copyright 2015, EUMETSAT, All Rights Reserved
 *
 * \par PRODUCED BY :  GMV Aerospace and Defence S.A.U.
 *                     This SW was developed by GMV within the
 *                     context of the EUMETSAT Satellite Application
 *                     Facility Co-operation Agreement for Support to
 *                     Nowcasting and Very Short Range Forecasting,
 *                     dated on 29 June 2011 between EUMETSAT
 *                     and AEMET.
 *
 * \par PROJECT     :  NWC SAF
 *
 * \par UNIT NAME   :  TOOLS
 *
 * \par FILE        :  GEO-L1SD.c
 *
 * \par TYPE        :  Source
 *
 * \par FUNCTION    :  
 * Lee bandas de satélite y genera ficheros temporales binarios o netCDF
 * 
 **************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 *  JUN 2016   1.0         GMV        Creación (Accion-RC-29-04)
 * 
 **************************************************************************/

#include "NWCLIB.h"
#include "nwcsat.h"
#include "nwcprod.h"
#include "netcdf.h"


/* Common constants and definitions */
#define PGE_ID   "GEO-L1SD"   /**< PGE Identifier (to be used in LOGs)    */
#define PGE_VERS "1.0"       /**< PGE Version (to be used in LOGs)       */

typedef enum { false, true } bool;
typedef struct {
  bool rad;       /**< Flag to produce band data in radiance */
  bool refl_bt;   /**< Flag to produce band data in reflectance/brightness temperature */
  bool refl_norm; /**< Flag to produce normaliced reflectances (if reflectances are requested) */
  bool binary;    /**< Flag to produce output products in binary format */
  bool netcdf;    /**< Flag to produce output products in binary format */
    
} L1sdConf;



/* Private Functions */
Status l1sd_initialise(char *region_conf_file, char *model_conf_file,
           Psing_region **region, char *band, U_int *n_bands, char ***band_name,
           L1sdConf *conf);
void l1sd_StrSplit(char *char_in, U_int *n, char ***out);
void l1sd_processSat(char *exe, Utc *slot, char *cfg, char *cfm, 
                 Psing_region *region, U_int n_bands, 
                 char **band_name,L1sdConf *conf);
Status l1sd_ProdWriteDataFloat(int ncid, Head_structure *head,  char *variable, 
                        Float_32 **data, char *band_name, Sat_cal cal);
Status l1sd_Rename_Product(Psing_region *region, Sat_res res, 
                           char *band_id, Sat_cal cal, Utc *slot);
Status l1sd_Remove_Binary(Psing_region *region, char *band_id, 
                          Sat_cal cal, Sat_res res, Utc *slot);
Status l1sd_ReadWrite(Psing_region *region, Sat_band band, char *band_name, 
               Sat_cal cal, Sat_res res, Utc *slot, Satellite_data *sat, 
               Head_structure *head, char *comment,L1sdConf *conf);



/****************************************************************************/
/**
 * \brief main  Lectura de bandas de satélite
 * 
 * \param[in] argc Command line arguments:<br>
 *         usage: \<executable\> YYYY-MM-DDThh:mm:ssZ region_conf_file model_conf_file
 *         <ul>
 *         <li><em>\<executable\></em> is the name of the PGE executable
 *         <li><em>YYYY-MM-DDThh:mm:ss</em> is the nominal time for the slot to be processed
 *         <li><em>region_conf_file</em> is the name of the file defining the region
 *         <li><em>model_conf_file</em> is the name of the file containing the
 *                                    model configurable parameters for the PGE
 *         </ul>
 * 
 * \param[in] argv Number of arguments<br>
 * 
 * \return         0 : OK      - Product Successfully generated<br>
 *          [  1-125]: Warning - Product generated but probably incomplete and/or
 *                               with reduced quality<br>
 *          [128-256]: Error   - Product is not generated<br>
 * 
 * \par DESCRIPTION: 
 * Esta herramienta lee productos de satélite y genera ficheros intermedios 
 * en formatos binario y netCDF para su uso posterior por otros programas
 * 
 *   
 * \par WARNINGS/ERRORS:
 *   ERROR_COMMAND:     Error in command line<br>
 *   ERROR*: Error code from:
 *           <ul>
 *           <li> l1sd_initialise()
 *           <li> l1sd_getHeight()
 *           <li> l1sd_doParallax()
 *           </ul>
 *
 *****************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 *  MAY 2015   1.0         GMV        Creation of GEO-PLAX PGE/Tool
 * 
 **************************************************************************/
int main (int argc, char *argv[]) {
    
  Utc           slot;
  Psing_region  *region=NULL;
  Status        status;
  
  char band[MAX_VALUE_STRING];
  char **band_name=NULL;
  U_int n_bands;
  
  L1sdConf conf;
  
  
  /* ------------------------------------------------------------------------ */
  /* Log the beginning of the process                                         */
  /* ------------------------------------------------------------------------ */
  NwcLog(I, PGE_ID, PGE_VERS, "Initializing PGE %s V%s",PGE_ID,PGE_VERS);
  
  /* ------------------------------------------------------------------------ */
  /* Read line arguments                                                      */
  /* ------------------------------------------------------------------------ */
  if(argc != 4) {
      NwcLog(E, PGE_ID, PGE_VERS, "Error Executing %s", argv[0]);
      NwcLog(E, PGE_ID, PGE_VERS, 
              "Usage: %s YYYY-MM-DDThh:mmZ <region_conf_file> <model_conf_file>",
              argv[0]);
      return ERROR_COMMAND;
  } else {
      NwcLog(I, PGE_ID, PGE_VERS, "%s %s %s %s",
              argv[0],argv[1],argv[2],argv[3]);
  }
  
  
  /* ------------------------------------------------------------------------ */
  /* Initializing                                                             */
  /* ------------------------------------------------------------------------ */
  // First set the Date/Slot in a Utc structure
  // -------------------------------------------
  if((status=NwcTimeSetStr(&slot,argv[1])) != OK) {
      NwcLog(E, PGE_ID, PGE_VERS, 
              "Error in date format (%s). Required YYYY-MM-DDThh:mmZ",argv[1]);
      return ERROR_COMMAND;
  }
  
  /* Initialize Data Structures */
  /* -------------------------- */
  if((status=l1sd_initialise(argv[2], argv[3], &region, 
                        band, &n_bands, &band_name,
                        &conf))!=OK) {
     NwcLog(E, PGE_ID, PGE_VERS, "Error in module l1sd_initialise()");
     return status;
  }
  
  /* ------------------------------------------------------------------------ */
  /* Processing the list of Satellite Bands to be parallax-corrected          */
  /* ------------------------------------------------------------------------ */
  l1sd_processSat(basename(argv[0]), &slot, argv[2], argv[3], region, 
                        n_bands,  band_name, &conf);
    
  /* ------------------ */
  /* Finalizing the PGE */
  /* ------------------ */
  // Deallocating memory
  NwcRegionFree(&region);
  free(band_name);
  
  NwcLog(I, PGE_ID, PGE_VERS, "Concluding PGE %s V%s",PGE_ID, PGE_VERS);

  return OK;
}

/*************************************************************************/
/**
 * 
 * \brief l1sd_initialise - Initialize the processing environment for the
 *                          execution of the GEO-L1SD
 *
 * \param[in] region_conf_file Region Configuration filename
 * \param[in] model_conf_file Model Configuration filename
 * \param[out] region Processing Region (allocated within the function)
 * \param[out] band List of Satellite bands to be read (from cfm)
 * \param[out] n_bands Number of Satellite bands to be read (from cfm)
 * \param[out] band_name Array with the list of Satellite bands ids to be read (from cfm)
 * \param[out] conf L1SD Configuration
 * 
 * 
 * \return OK: Process concluded successfully<br>
 *   ERROR*: Error code from:
 *           <ul>
 *           <li> NwcRegionSet() (Error initializing the Processing Region)
 *           <li> NwcCFRead() (Error reading the Model Configuration File)
 *           </ul>
 * 
 * \par DESCRIPTION:
 * This function initialize the processing environment for the execution of
 * the GEO-L1SD:
 * <ul>
 * <li> Set the processing region
 * <li> Read the Model configuration file parameters
 * </ul>
 * 
 **************************************************************************/
/* HISTORY
 *
 * DATE      VERSION       AUTHOR     REASONS
 *  JUN 2016   1.0         GMV        Creation of GEO-L1SD PGE/Tool
 * 
 **************************************************************************/
Status
l1sd_initialise(char *region_conf_file, char *model_conf_file,
           Psing_region **region, 
           char *band, U_int *n_bands, char ***band_name,
           L1sdConf *conf)
{    
  Status status;    
  Conf_param_list param_list=NULL;
  char aux[MAX_VALUE_STRING];
  
  
  NwcLog(I, PGE_ID, PGE_VERS, "Executing l1sd_initialise() ...");
    
  /* ------------------------------------------------------------------ */
  /* Initialize the data structure defining the region  to be processed */
  /* ------------------------------------------------------------------ */
  if((status=NwcRegionSet(region_conf_file, region)) != OK) {
    NwcLog(E, PGE_ID, PGE_VERS, "Error setting the region from %s ", region_conf_file);
    return status;
  }
  
  /* ------------------------------------ */
  /* Read parameters from model conf file */
  /* ------------------------------------ */
  if((status=NwcCFRead(model_conf_file, &param_list))!=OK) {
    NwcLog(E, PGE_ID, PGE_VERS, "Error reading %s",model_conf_file);
    NwcRegionFree(region);
    return status;
  }
  
  // BANDS
  // ---------
  if(NwcCFGetStr(param_list,"L1SD_BANDS",band)!=OK) band[0]='\0';
  NwcLog(I, PGE_ID, PGE_VERS, "Satellite Bands to be read: >%s<",band);
  l1sd_StrSplit(band, n_bands, band_name);

  // RAD and REFL_BT
  // ----------------
  conf->rad       = false;
  conf->refl_bt   = false;
  conf->refl_norm = false;
  
  NwcCFGetStr(param_list,"L1SD_RAD",aux);
  if(strcmp(aux,"YES") == 0) conf->rad=true;
  
  NwcCFGetStr(param_list,"L1SD_REFL_BT",aux);
  if(strcmp(aux,"YES") == 0) conf->refl_bt=true;
  
  NwcCFGetStr(param_list,"L1SD_REFL_NORM",aux);
  if(strcmp(aux,"YES") == 0) conf->refl_norm=true;

  // BINARY and NETCDF outputs
  // -------------------------
  conf->binary = false;
  conf->netcdf = false;
  
  NwcCFGetStr(param_list,"L1SD_BINARY",aux);
  if(strcmp(aux,"YES") == 0) conf->binary=true;

  NwcCFGetStr(param_list,"L1SD_NETCDF",aux);
  if(strcmp(aux,"YES") == 0) conf->netcdf=true;
  

  // Closing the Configuration file
  NwcCFFree(&param_list);
  
  return OK;
}


/****************************************************************************/
/**
 * \brief l1sd_StrSplit - Split a string in an array of words 
 * 
 * \param[in,out] char_in Input string<br>
 *                        On output, string splitted in tokens by strtok()
 * \param[out] n Number of words in char_in
 * \param[out] out array of char pointers storing the list of words in char_in.
 *             <b>Note:</b> The array is allocated by within the function
 *
 * \return void
 * 
 * \par DESCRIPTION: 
 *  This function splits an input string in words. The individual words are
 *  stored in an array of (char *) allocated within the function. Note that the 
 *  array of pointers point to positions of the input string char_in splitted in
 *  tokens by strtok(). Therfore, input array is modified by the functions, 
 *  and the output array of pointers (char *) simply points to positions
 *  in char_in.
 *   
 *****************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 *  JUN 2016   1.0         GMV        Creation of GEO-PLAX PGE/Tool
 * 
 **************************************************************************/
void 
l1sd_StrSplit(char *char_in, U_int *n, char ***out)
{
    char *aux1, *aux2;
    char **c=NULL;
    U_int i;
    
    // First count the number of elements in char_in
    // A copy is used to avoid modifying input string (for further processing)
    aux1=strdup(char_in);
    *n=0;
    aux2=strtok(aux1," ,");
    while(aux2 != NULL)   {
       (*n) ++;
       aux2=strtok(NULL, " ,");
    }
    free(aux1);
    
   // Allocate memory and store the pointer to each element
   // Is an array of pointer to the original char_in, splitted in tokens
   if(*n > 0) {
       c = (char **) malloc(*n * sizeof (char*));
       c[0]=strtok(char_in," ,");
       for(i=1; i < *n; i++) c[i]=strtok(NULL, " ,");
   }
    
   *out=c;
}


/*************************************************************************/
/**
 * 
 * \brief l1sd_processSat - Read/Store Satellite data
 *
 * \param[in] exe Name of the executable
 * \param[in] slot Processing slot
 * \param[in] cfg Region Configuration filename
 * \param[in] cfm Model Configuration filename
 * \param[in] region Processing Region
 * \param[in] n_bands Number of bands to be corrected
 * \param[in] band_name Array with the list of band names to be corrected
 * \param[in] conf L1SD Configuration
 * 
 * \return void
 * 
 * \par DESCRIPTION:
 * This function parallax-correct a Satellite Product:
 * <ul>
 * <li> Open the Satellite data
 * <li> Perform the parallax correction in all user selected satellite bands
 * <li> Writes the new PLAX (satellite data) product 
 * </ul>
 * 
 **************************************************************************/
/* HISTORY
 *
 * DATE      VERSION       AUTHOR     REASONS
 * JUN 2016    1.0         GMV        Creation of GEO-PLAX PGE/Tool
 * 
 **************************************************************************/
void 
l1sd_processSat(char *exe, Utc *slot, char *cfg, char *cfm, 
                 Psing_region *region, U_int n_bands, char **band_name,
                 L1sdConf *conf)
{
  U_int i;
  Head_structure head;
  Sat_band band;
  Satellite_data sat;
  Sat_res res;
    
    
  if(n_bands != 0) {  // Only process if bands are requested
      
    // Open the Satellite input data
    NwcLog(I, PGE_ID, PGE_VERS,  "Opening Satellite Data for slot %s",slot->str);
    if(NwcSatInit(region, slot, &sat) == OK)  { // If Sat data are available
        
        // Process each required band
        for(i=0; i<n_bands; i++) { // loop to all bands in RSB_BAND
            
            // Get band_id from band string
            if((band=nwcSatBandToEnum(band_name[i])) == 255) {
              NwcLog(E, PGE_ID, PGE_VERS,  "Satellite band %s not known", band_name[i]);
              continue;
            }
            res = region->sat.rad[band].res; // Nominal resolution for band
          
            /* Common Header data */
            /* ------------------ */
            strcpy(head.exe, exe);  // Name of the executable
            strcpy(head.cfg, cfg); // Name of the configuration file
            strcpy(head.cfm, cfm); // Name of the model configuration file
            strcpy(head.pge_vers,PGE_VERS);       // Version of the PGE
            head.product_completeness = 100;
            head.product_quality      = 100;


            // Reading the band - RADIANCES
            // ----------------------------
            if(conf->rad == true)                 
              l1sd_ReadWrite(region, band, band_name[i], RAD, res, slot, &sat,
                             &head, "RADIANCE", conf);
            

            // Reading the band - REFLECTANCES/BT
            // ----------------------------------
            if(conf->refl_bt == true) {
              if(region[res].sat.rad[band].vc == 0) {  // NON-THERMAL CHANNEL
                  if(conf->refl_norm == true) 
                    l1sd_ReadWrite(region, band, band_name[i], REFN, res, slot, &sat,
                             &head, "NORMALIZED REFLECTANCE", conf);
                  else 
                    l1sd_ReadWrite(region, band, band_name[i], REFL, res, slot, &sat,
                             &head, "REFLECTANCE", conf);
              }
              else { // THERMAL CHANNEL
                    l1sd_ReadWrite(region, band, band_name[i], BT, res, slot, &sat,
                             &head, "BRIGHTNESS TEMPERATURE", conf);
                   }

            }  // end if refl/bt are requested
        } // endloop all bands
        
        /* Closing input and output PGE */
        /* ---------------------------- */
        NwcSatFreeBand(&sat,NUM_BANDS,NUM_CALS); // To deallocate all memory
    }
    else { // Unable to open Satellite Data
        NwcLog(E, PGE_ID, PGE_VERS,  "Error initiating satellite data for slot %s", slot->str);
    }
      
  }  // end if n_bands = 0 
}



/****************************************************************************/
/**
 * \brief l1sd_ProdWriteDataFloat - Writes a Float Container in
 *                            a NWC/GEO product in netCDF format 
 * 
 * \param[in] ncid Product file id obtained by NwcProdWriteInit()
 * \param[in] head Header Structure initiated by NwcProdWriteInit()
 * \param[in] variable Name of the float variable to be 
 *                     added to the NWG/GEO product
 * \param[in] data 2D array containing the float values in 
 *                 native units
 * \param[in] band_name Band identifier
 * \param[in] cal Selected calibration
 * 
 * \return OK: Container has been successfully written in the NWC/GEO product<br>
 *         ERROR: Error in netCDF functions
 * 
 * 
 * \par DESCRIPTION:
 * This function writes a Float container named
 * <em>variable</em> in a NWC/GEO output initialized by NwcProfWriteInit().<br>
 * <br>
 * The Float values are provided as Float_32 in native units.<br>
 * <br>
 * Values out of <em>valid_range</em> 
 * will be stored as <em>valid_range[0]</em> (values below the range) or 
 * <em>valid_range[1]</em> (values above the range).<br>
 * <br>
 * Values set to NODATA in <em>data</em> array will be coded as 
 * <em>_FillValue</em> within the netCDF file.
 *
 *****************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 * JUN 2016    DRR         GMV      First Version of the GEO-L1SD
 *
 ****************************************************************************/
Status 
l1sd_ProdWriteDataFloat(int ncid, Head_structure *head, char *variable, 
                        Float_32 **data, char *band_name, Sat_cal cal)
{
    int dimid[2];
    int varid;
    int nc_status;
    Conf_param_list param_list=NULL;
    Conf_param_list block_list=NULL;
    char prodio[MAX_VALUE_STRING];
    Status status;
    Float_32 valid_range[2], fill_value;
    char key_init[MAX_VALUE_STRING];
    char long_name[MAX_VALUE_STRING];
    char units[MAX_VALUE_STRING];
    U_int i,j;
    size_t dimlen[2];

    // --------------------------------------------------
    // Getting the name of the PRODIO Configuration file
    // --------------------------------------------------
    sprintf(prodio,"%s/%s/%s/%s_%s.cf",getenv(SAFNWC_HOME_ENV_VAR), AUX_DIR, head->prodname, 
            PRODIO_CONF_FILENAME_PRIVATE, head->prodname);
    if((status=NwcCFRead(prodio, &param_list)) != OK) {
       NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
          "l1sd_ProdWriteDataFloat: Unable to read the PRODIO Configuration File %s",
           prodio);
       return status;
    }
    
    // -----------------------------------------------------
    // Retrieving the dimension ids and dimensions
    // -----------------------------------------------------
    if((nc_status=nc_inq_dimid(ncid, "ny", &dimid[0])) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to define the dimension ny in %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_inq_dimid Error: %s",nc_strerror(nc_status));
        NwcCFFree(&param_list);
        return ERROR_FILE_IO;
    }
    if((nc_status=nc_inq_dimlen(ncid, dimid[0], &dimlen[0])) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to retrieve the dimension length for ny in %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_inq_dimlen Error: %s",nc_strerror(nc_status));
        return ERROR_FILE_IO;
    }
    
    if((nc_status=nc_inq_dimid(ncid, "nx", &dimid[1])) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to define the dimension nx in %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_inq_dimid Error: %s",nc_strerror(nc_status));
        NwcCFFree(&param_list);
        return ERROR_FILE_IO;
    }
    if((nc_status=nc_inq_dimlen(ncid, dimid[1], &dimlen[1])) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to retrieve the dimension length for nx in %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_inq_dimlen Error: %s",nc_strerror(nc_status));
        return ERROR_FILE_IO;
    }

    
    // -------------------------------
    // Creating the Variable container
    // -------------------------------
    if((nc_status=nc_def_var(ncid, variable, NC_FLOAT, 2, dimid, &varid)) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to create the Variable container %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_def_var Error: %s",nc_strerror(nc_status));
        NwcCFFree(&param_list);
        return ERROR_FILE_IO;
    }
    nc_def_var_deflate(ncid, varid, NC_SHUFFLE, 1, DEFLATE_LEVEL);

    // ----------------------------------------
    // Writing the  Attributes of the Container
    // First retrieve the applicable block from the configuration file
    // ----------------------------------------
      sprintf(long_name,"l1_satellite_data_band_%s_in_",band_name);
      if(cal==RAD) {
          strcat(long_name,"radiance");
          sprintf(units,"mW m-2 sr-1 (cm-1)-1");
      }
      if(cal==REFL) {
          strcat(long_name,"reflectance");
          sprintf(units, "%%");
      }
      if(cal==REFN) {
          strcat(long_name,"normalized_reflectance");
          sprintf(units, "%%");
      }
      if(cal==BT) {
          strcat(long_name,"brightness_temperature");
          sprintf(units, "K");
      }

     sprintf(key_init,"BLOCK_%s",variable);
    if(NwcCFGetBlock(param_list, key_init, "BLOCK_END", &block_list) == OK) {
      nwcProdWriteAttTxtFromCF(ncid, varid, "standard_name", block_list);
//      nwcProdWriteAttTxtFromCF(ncid, varid, "long_name", block_list);
//      nwcProdWriteAttTxtFromCF(ncid, varid, "units", block_list);
      nwcProdWriteAttTxt(ncid, varid, "long_name", long_name);
      nwcProdWriteAttTxt(ncid, varid, "units", units);
      nwcProdWriteAttFloatFromCF (ncid, varid, "valid_range", 2, block_list);
      nwcProdWriteAttFloatFromCF (ncid, varid, "_FillValue", 1, block_list);
      nwcProdWriteAttTxtFromCF(ncid, varid, "ancillary_variables", block_list);
      nwcProdWriteAttTxtFromCF(ncid, varid, "coordinates", block_list);
      nwcProdWriteAttTxtFromCF(ncid, varid, "comment", block_list);
  
      
    }
    else
      NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
           "l1sd_ProdWriteDataFloat: Unable to read block %s in %s",key_init, prodio);

    // -------------------------
    // Finally, writing the data
    // -------------------------
    nc_enddef(ncid);
    
    // -------------------------------------------------------------------
    // Check for values (valid_range and _FillValue)
    // -------------------------------------------------------------------
    if(NwcCFGetFloat(block_list, 2, "valid_range", valid_range) != 2) {
       NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
          "l1sd_ProdWriteDataFloat: Unable to read the valid_range value from the "
           "PRODIO Configuration File %s", prodio);
       NwcCFFree(&param_list);
       NwcCFFree(&block_list);
       return ERROR_FILE_IO;
    }

    if(NwcCFGetFloat(block_list, 1, "_FillValue", &fill_value) != 1) {
       NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
          "l1sd_ProdWriteDataFloat: Unable to read the _FillValue value from the "
           "PRODIO Configuration File %s", prodio);
       NwcCFFree(&param_list);
       NwcCFFree(&block_list);
       return ERROR_FILE_IO;
    }
    
    NwcCFFree(&block_list);
    
    
    // Check the data
    for(i=0; i<dimlen[0]; i++) {
        for(j=0; j<dimlen[1]; j++) {
            if(data[i][j] == NODATA) {
                data[i][j]=fill_value;
            } else {
              if(data[i][j] < valid_range[0]) data[i][j]=valid_range[0];
              if(data[i][j] > valid_range[1]) data[i][j]=valid_range[1];
            }
        }
    }
    
    // Store the data
    if((nc_status=nc_put_var_float(ncid,varid,data[0])) != NC_NOERR) {
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: Unable to store data in container %s", variable);
        NwcLog(E,NWCLIB_ID, NWCLIB_VERS,
            "l1sd_ProdWriteDataFloat: nc_put_var_float Error: %s",nc_strerror(nc_status));
        NwcCFFree(&param_list);
        return ERROR_FILE_IO;
    }    
    
   NwcCFFree(&param_list);
   return(OK);
}

/****************************************************************************/
/**
 * \brief l1sd_Rename_Product - Renames L1Sd product generated by the NWCLIB, 
 *                              adding band and calibration
 * 
 * \param[in] region Processing region
 * \param[in] res Satellite resolution
 * \param[in] band_name Band identifier
 * \param[in] cal Selected calibration
 * \param[in] slot Nominal slot time
 * 
 * \return OK: Product has been succesfully renamed
 * 
 * 
 * \par DESCRIPTION:
 * This function renames a L1Sd product generated by the NWCLIB using the
 * naming criteria defined in the DOF to include the band and the calibration
 *
 *****************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 * JUN 2016    DRR         GMV      First Version of the GEO-L1SD
 *
 ****************************************************************************/
Status 
l1sd_Rename_Product(Psing_region *region, Sat_res res, char *band_name, 
                    Sat_cal cal, Utc *slot)
{
    char filename[MAX_VALUE_STRING];
    char filepath[MAX_VALUE_STRING];
    char fullname1[MAX_VALUE_STRING];
    char fullname2[MAX_VALUE_STRING];
    char command[MAX_VALUE_STRING];
    char date_str[20];
    char cal_str[20];

                  
    NwcProdGetPath("L1SD", filepath);
    NwcProdGetName("L1SD", region, res, slot, 0, filename);
    sprintf(fullname1,"%s/%s", filepath, filename);   
    
    // Date representation in output product filename as YYYYMMDDThhmmssZ
    sprintf(date_str, "%04d%02d%02dT%02d%02d%02dZ",
            slot->tm.tm_year+1900, slot->tm.tm_mon+1, slot->tm.tm_mday,
            slot->tm.tm_hour, slot->tm.tm_min, slot->tm.tm_sec);
    if(cal == RAD)         sprintf(cal_str,"RAD");
    else if (cal == REFL)  sprintf(cal_str,"REFL");
    else if (cal == REFN)  sprintf(cal_str,"REFN");
    else if (cal == BT)    sprintf(cal_str,"BT");
    
    sprintf(filename,"S_NWC_%s-%s_%s_%s-%s_%s.nc",
                 band_name, cal_str, region[res].satellite, region[res].id, region[res].res_id,
                 date_str);
    sprintf(fullname2,"%s/%s", filepath, filename);   
    
    sprintf(command, "mv %s %s",fullname1, fullname2);
    system(command);
    
    return OK;
    
}

/****************************************************************************/
/**
 * \brief l1sd_Remove_Binary - Removes a specific binary product
 * 
 * \param[in] region Processing region
 * \param[in] band_name Band identifier
 * \param[in] cal Selected calibration
 * \param[in] res Satellite Resolution
 * \param[in] slot Nominal slot time
 * 
 * \return OK: Product has been successfully removed
 * 
 * 
 * \par DESCRIPTION:
 * This function removes a specific binary file
 *
 *****************************************************************************/
/* HISTORY
 *
 *  DATE      VERSION     AUTHOR     REASONS
 * JUN 2016    DRR         GMV      First Version of the GEO-L1SD
 *
 ****************************************************************************/
Status l1sd_Remove_Binary(Psing_region *region, char *band_name, 
                          Sat_cal cal, Sat_res res, Utc *slot)
{
    
  char home[MAX_VALUE_STRING];
  char file[MAX_VALUE_STRING];
  char command[MAX_VALUE_STRING];
    
  if(getenv(SAFNWC_HOME_ENV_VAR) == NULL) {
    NwcLog(E, NWCLIB_ID, NWCLIB_VERS,
                "l1sd_Remove_Binary: SAFNWC environment variable is not set");
    return ERROR;
  }
  sprintf(home,"%s/%s",getenv(SAFNWC_HOME_ENV_VAR), DATABUF_DIR);

  sprintf(file, "%s/S_NWC_%s_%s_%s-%s_%s",
            home, band_name, region[res].satellite, 
            region[res].id, region[res].res_id, slot->str);


  if(cal == RAD)       strcat(file,".rad");
  else if(cal == REFL) strcat(file,".refl");
  else if(cal == REFN) strcat(file,".refn");
  else if(cal == REFLC) strcat(file,".reflc");
  else if(cal == REFNC) strcat(file,".refnc");
  else if(cal == BT)   strcat(file,".bt");
  
  sprintf(command,"rm -f %s",file);
  system(command);
  
  return OK;
    
}        
            
/*************************************************************************/
/**
 * 
 * \brief l1sd_ReadWrite - Read/Store Satellite data
 *
 * \param[in] region Processing Region
 * \param[in] band band identifier (enum)
 * \param[in] band_name band identifier (*char)
 * \param[in] cal Selected calibration
 * \param[in] res Satellite Resolution
 * \param[in] slot Nominal Slot time
 * \param[in] sat Satellite data
 * \param[in] head Output product header structure
 * \param[in] comment Comment to include in logs
 * \param[in] conf L1SD Configuration
 * 
 * \return void
 * 
 * \par DESCRIPTION:
 * This function reads and writes satellite data in the defined region, in 
 * selected calibration y resolution, and writes output data in binary and/or
 * netCDF according the configuration
 * 
 **************************************************************************/
/* HISTORY
 *
 * DATE      VERSION       AUTHOR     REASONS
 * JUN 2016    1.0         GMV        Creation of GEO-PLAX PGE/Tool
 * 
 **************************************************************************/
Status            
l1sd_ReadWrite(Psing_region *region, Sat_band band, char *band_name, 
               Sat_cal cal, Sat_res res, Utc *slot, Satellite_data *sat,
               Head_structure *head, char *comment,
               L1sdConf *conf)
{
   int ncid_out;
   
  NwcLog(I, PGE_ID, PGE_VERS,  "Reading Band %s for slot %s - %s", 
                               band_name, slot->str, comment);
  if(NwcSatReadBand(region, band, cal, res, sat) != OK) {
    NwcLog(E, PGE_ID, PGE_VERS,  "Unable to read Satellite band %s", band_name);
    return ERROR;
  }

  // Writting the product in netCDF
  if(conf->netcdf==true) {
    NwcProdWriteInit("L1SD", region, res, slot, 0, head, &ncid_out);
    nwcProdWriteAttTxt(ncid_out, NC_GLOBAL, "processing_level", "Level 1");

    if(cal==RAD) l1sd_ProdWriteDataFloat(ncid_out, head,  "data", sat->rad[band].rad, band_name, cal);
    else         l1sd_ProdWriteDataFloat(ncid_out, head,  "data", sat->rad[band].bt, band_name, cal);
    NwcProdWriteClose(ncid_out, head, region, res, NODATA, NODATA, NULL, NULL);
    l1sd_Rename_Product(region, res, band_name, cal, slot);
  }

  // Removing Binary product if not required
  if(conf->binary==false) 
      l1sd_Remove_Binary(region, band_name, cal, res, slot);
  
  return OK;
  
}
